Some Title

# A
{% md_clean_include 'sub_content_with_subs.md' heading_level=2 %}

# B
{% if variables.mac_os %}
Users are recommended to use homebrew to install the latest python 3 and then install python within a venv.
{% endif %}