# Control Page

{{ control_writer.write_control_with_sections(control,
profile,
group_title,
['objective', 'ImplGuidance', 'above_the_line_guidance', 'expected_evidence'],
{
'objective':'Control Objective Header',
'ImplGuidance': "The implementation guidance",
'above_the_line_guidance': "The above the line guidance",
'add_to_part_a': 'Add to part a',
'evidence_guidance': "Evidence Guidance"
},
label_column=True,
add_group_to_title=True
)
| safe
}}
