Test SSP
<!--
1. Adjust md_clean_include (or make it md_adjust_include) to adjust max header level
2. (CB) - Add ssp_renderer to jinja.py vars
3. Decide if you want just Values column or separate 'Values' and 'Label or Choices' by setting false to true for table
-->

# A
{% md_clean_include 'sub_content.md' heading_level=2 %}

# B
{% md_clean_include 'sub_content.md' heading_level=2 %}

# C
{% md_clean_include 'sub_content.md' heading_level=2 %}

# Control Response
{% for group in catalog_interface.get_all_groups_from_catalog() +%}
{% if group.class != '' %}
## {{ group.title }} {{ group.class }} \({{ group.id|upper }}\)
{% else %}
## {{ group.title }} \({{ group.id|upper }}\)
{% endif %}

{% for control in catalog_interface.get_sorted_controls_in_group(group.id) %}
{# ### {{ control_interface.get_label(control) }}: {{ control.title}}

#### Control Statement #}
{{ ssp_md_writer.get_control_statement(control.id, 2) }}

#### {{ control_interface.get_label(control) }} Summary information
{{ ssp_md_writer.get_fedramp_control_tables(control.id, 4, false) }}

#### Implementation
{{ ssp_md_writer.get_control_response(control.id, 4)}}

{% endfor %}
{% endfor %}
